(* file: rewrite.sml *)
(* description: rewriting *)
(* author: Takahito Aoto *)

signature REWRITE = 
sig 
    val rootStep: Trs.trs -> Term.term -> Term.term option
    val step: Trs.trs -> Term.term -> Term.position -> Term.term option
    val listep: Trs.trs -> Term.term -> Term.term option
    val isNF: Trs.trs -> Term.term -> bool
    val isNF_root: Trs.trs -> Term.term -> bool		  
    val listepsToNF: Trs.trs -> Term.term -> unit
    val linf: Trs.trs -> Term.term -> Term.term
    val isJoinable: Trs.trs -> Term.term * Term.term -> bool
end

structure Rewrite: REWRITE =
struct 

local 
    structure L = List
    structure S = Subst
    structure T = Term
in

(* 最内最左の順に並んだpositionのリストを返す *)
fun posLeftInMost t =
    let
	(* 組(x,v)のリストのvの最大値を返す *)
	fun max xs = foldl (fn ((x,v), mav) => if v > mav then v else mav) 0 xs 
	(* 組(x,v)のリストPとvの最大値mavを受け取り，vの降順にxを並び替える．*)
	(* ただし，vが等しいものはPの順番通りに並ぶ *)
	fun sort xs =
	    let fun filter xs n = foldr (fn ((x,v),(Eq,Gr)) => if v = n then (x::Eq,Gr) else (Eq,(x,v)::Gr))
					([],[]) xs
		fun main [] n = []
		  | main xs n =
		    let val (Eq,Gr) = filter xs n 
		    in
			Eq @ (main Gr (n-1))
		    end
	    in
		main xs (max xs)
	    end
    in sort (Term.posWithDepth t)
    end
	
(* rootStep rs term : TRS rs による根位置での書き換え *)
fun rootStep [] term = NONE
  | rootStep ((l,r)::rs) term = 
    case S.match l term
     of SOME sigma => SOME (S.apply sigma r)
      | NONE => rootStep rs term

(* step rs term ps : TRS rs による位置psでの書き換え *)
fun step rs term ps = let val rsigmaOp = rootStep rs (T.subterm term ps)
			  val c = T.makeContext term ps
		      in
			  case rsigmaOp of
			      NONE => NONE
			    | SOME rsigma => SOME (T.fillContext c rsigma)
		      end
	  
(* listep rs term : TRS rsによる最左最内書き換え *)
fun listep rs term = 
    let fun main rs [] = NONE
	  | main rs (p::ps) = case (step rs term p) of
				  NONE => main rs ps
			       |  SOME t => SOME t 
    in
	main rs (posLeftInMost term)
    end
	
fun isNF rs (T.Var x) = true
  | isNF rs (T.Fun (f,ts)) = case rootStep rs (T.Fun (f,ts)) of
				 SOME t' => false
			       | NONE => List.all (isNF rs) ts

(* NFかをチェックする．rootの位置だけをチェックする． *)
fun isNF_root rs (T.Var x) = true
  | isNF_root rs (T.Fun (f,ts)) = case rootStep rs (T.Fun (f,ts)) of
				 SOME t' => false
			       | NONE => true
						  
fun listepsToNF rs term =
  let fun main term = case (listep rs term) of
			  NONE => ()
			| SOME t => (print ("->R "^Term.toString t^"\n"); main t)
  in
      (print ("    "^Term.toString term^"\n"); main term)
  end

fun linf rs (T.Var x) = T.Var x
  | linf rs (T.Fun (f,ts)) = let val ts' = map (fn t => linf rs t) ts
			     in case rootStep rs (T.Fun (f,ts')) of
				    SOME t' => linf rs t'
				  | NONE => T.Fun (f,ts')
			     end

fun isJoinable rs (t1,t2) = let val t1' = linf rs t1
				val t2' = linf rs t2
			    in
				t1' = t2'
			    end
					    
end (* of local *)

end
