(* file: trs.sml *)
(* description: first-order term rewriting systems *)
(* author: Takahito Aoto *)

signature TRS = 
sig
    type rule = (Term.term * Term.term)
    type eq = (Term.term * Term.term)
    type trs = rule list
    type eqs = eq list

    val prRule: rule -> string
    val rdRule: string -> rule

    val prRules: trs -> string
    val rdRules: string -> trs

    val prEq: eq -> string
    val rdEq: string -> eq

    val prEqs: eqs -> string
    val rdEqs: string -> eqs

    val rename: (Term.term * Term.term) * (Term.term * Term.term)
		-> (Term.term * Term.term) * (Term.term * Term.term)

    val filter: Fun.key -> trs -> trs
end

structure Trs: TRS =
struct

local 
    structure SU = StringUtil
    structure T = Term
    structure LU = ListUtil
in
type rule = (Term.term * Term.term)
type eq = (Term.term * Term.term)
type trs = rule list
type eqs = eq list

fun prRule (l,r) = T.toString l ^ " -> "  ^ T.toString r
fun rdRule str = T.readKeySeparatedTermPair "->" str 

fun prRules rs = LU.toStringCommaLnSquare prRule rs
fun rdRules str = T.readMultipleKeySepartedTermPairs ("[",",","]") "->" str

fun prEq (l,r) = T.toString l ^ " = "  ^ T.toString r
fun rdEq str = T.readKeySeparatedTermPair "=" str 

fun prEqs es = LU.toStringCommaLnSquare prEq es
fun rdEqs str = T.readMultipleKeySepartedTermPairs ("[",",","]") "=" str

fun rename ((l1,r1), (l2,r2)) =
  let (* tのインデックスの最大値を計算する *)
      fun maxIndexOfTerm t =
	    foldl (fn (x,m) => Int.max (Var.index x, m)) 0 (Term.vars t)
      (* tの各変数のインデックスにそれぞれnを足す *)
      fun addIndex (Term.Var x) n = Term.Var (Var.name x, (Var.index x) + n)
	| addIndex (Term.Fun (f,ts)) n = Term.Fun (f, addIndexList ts n)
      and addIndexList [] n = []
	| addIndexList (t::ts) n = (addIndex t n) :: addIndexList ts n
      val maxIndex = Int.max (maxIndexOfTerm l1, maxIndexOfTerm r1)
      val l2' = addIndex l2 (maxIndex + 1)
      val r2' = addIndex r2 (maxIndex + 1)
  in 
      ((l1,r1), (l2',r2'))
  end

fun filter f rs = List.filter (fn (T.Fun(g,ts),r) => f = g
				 | (_,r) => false) rs

end (* of local *)

end


